#include <J3ML/Geometry/LineSegment2D.hpp>

void Line2DClosestPointLineLine(const Vector2& v0, const Vector2& v10, const Vector2& v2, const Vector2& v32, float& d, float& d2) // TODO: Move to Line2D when that exists
{
    // assert(!v10.IsZero());
    // assert(!v32.IsZero());
    Vector2 v02 = v0 - v2;
    float d0232 = v02.Dot(v32);
    float d3210 = v32.Dot(v10);
    float d3232 = v32.Dot(v32);
    // assert(d3232 != 0.f); // Don't call with a zero direction vector.
    float d0210 = v02.Dot(v10);
    float d1010 = v10.Dot(v10);
    float denom = d1010*d3232 - d3210*d3210;

    d = (denom !=0) ? (d0232*d3210 - d0210*d3232) / denom : 0;

    d2 = (d0232 + d * d3210) / d3232;
}

Geometry::LineSegment2D::LineSegment2D(const Vector2 &a, const Vector2 &b)
        : A(a), B(b) {}

Vector2 Geometry::LineSegment2D::GetPoint(float d) const {
    return (1.f - d) * A + d * B;
}

Vector2 Geometry::LineSegment2D::CenterPoint() const {
    return (A + B) * 0.5f;
}

Vector2 Geometry::LineSegment2D::Centroid() const {
    return CenterPoint();
}

void Geometry::LineSegment2D::Reverse() {
    Swap(A, B);
}

Vector2 Geometry::LineSegment2D::Dir() const {
    return (B - A).Normalized();
}

Vector2 Geometry::LineSegment2D::AnyPointFast() const { return A; }

Vector2 Geometry::LineSegment2D::ExtremePoint(const Vector2 &direction) const {
    return Vector2::Dot(direction, B - A) >= 0.f ? B : A;
}

Vector2 Geometry::LineSegment2D::ExtremePoint(const Vector2 &direction, float &projectionDistance) const {
    Vector2 extremePoint = ExtremePoint(direction);
    projectionDistance = extremePoint.Dot(direction);
    return extremePoint;
}

void Geometry::LineSegment2D::Translate(const Vector2 &offset) {
    A += offset;
    B += offset;
}

void Geometry::LineSegment2D::Transform(const Matrix3x3 &transform) {
    A = transform.Mul(A);
    B = transform.Mul(B);
}

void Geometry::LineSegment2D::Transform(const Matrix4x4 &transform) {
    A = transform.Mul(A);
    B = transform.Mul(B);
}

void Geometry::LineSegment2D::Transform(const Quaternion &transform) {
    //A = transform.Mul(A);
    //B = transform.Mul(B);
}

float Geometry::LineSegment2D::Length() const { return A.Distance(B); }

float Geometry::LineSegment2D::LengthSq() const { return A.DistanceSq(B); }

float Geometry::LineSegment2D::LengthSquared() const { return LengthSq(); }

bool Geometry::LineSegment2D::IsFinite() const { return A.IsFinite() && B.IsFinite(); }

bool Geometry::LineSegment2D::Equals(const Geometry::LineSegment2D &rhs, float epsilon) const {
    return (A.Equals(rhs.A, epsilon) && B.Equals(rhs.B, epsilon) ||
            A.Equals(rhs.B, epsilon) && B.Equals(rhs.A, epsilon));
}

bool Geometry::LineSegment2D::Contains(const Vector2 &point, float epsilon) const {
    return ClosestPoint(point).DistanceSq(point) <= epsilon;
}

bool Geometry::LineSegment2D::Contains(const Geometry::LineSegment2D &rhs, float epsilon) const {
    return Contains(rhs.A, epsilon) && Contains(rhs.B, epsilon);
}

Vector2 Geometry::LineSegment2D::ClosestPoint(const Vector2 &point) const {
    float d;
    return ClosestPoint(point, d);
}

Vector2 Geometry::LineSegment2D::ClosestPoint(const Vector2 &point, float &d) const {
    Vector2 dir = B - A;
    d = J3ML::Math::Clamp01(Vector2::Dot(point - A, dir) / dir.LengthSquared());
    return A + d * dir;
}

Vector2 Geometry::LineSegment2D::ClosestPoint(const Geometry::LineSegment2D &other) const {
    float d, d2;
    return ClosestPoint(other, d, d2);
}

Vector2 Geometry::LineSegment2D::ClosestPoint(const Geometry::LineSegment2D &other, float &d) const {
    float d2; return ClosestPoint(other, d, d2);
}

Vector2 Geometry::LineSegment2D::ClosestPoint(const Geometry::LineSegment2D &other, float &d, float &d2) const {
    Vector2 dir = B - A;
    Line2DClosestPointLineLine(A, B - A, other.A, other.B - other.A, d, d2);


    return Vector2::Zero;
}

float Geometry::LineSegment2D::Distance(const Vector2 &point) const { float d; return Distance(point, d); }

float Geometry::LineSegment2D::Distance(const Vector2 &point, float &d) const {
    /// See Christer Ericson's Real-Time Collision Detection, p. 130.
    Vector2 closestPoint = ClosestPoint(point, d);
    return closestPoint.Distance(point);
}

float Geometry::LineSegment2D::Distance(const Geometry::LineSegment2D &other) const { float d, d2; return Distance(other, d, d2);}

float Geometry::LineSegment2D::Distance(const Geometry::LineSegment2D &other, float &d) const { float d2; return Distance(other, d, d2); }

float Geometry::LineSegment2D::Distance(const Geometry::LineSegment2D &other, float &d, float &d2) const {
    ClosestPoint(other, d, d2);
    return GetPoint(d).Distance(other.GetPoint(d2));
}

float Geometry::LineSegment2D::DistanceSq(const Geometry::LineSegment2D &other) const {
    float d, d2;
    ClosestPoint(other, d, d2);
    return GetPoint(d).DistanceSq(other.GetPoint(d2));
}

float Geometry::LineSegment2D::DistanceSq(const Vector2 &point) const {
    float d;
    /// See Christer Ericson's Real-Time Collision Detection, p.130.
    Vector2 closestPoint = ClosestPoint(point, d);
    return closestPoint.DistanceSq(point);
}

bool Geometry::LineSegment2D::Intersects(const Geometry::LineSegment2D &lineSegment, float epsilon) const {
    return Distance(lineSegment) <= epsilon;
}

void Geometry::LineSegment2D::ProjectToAxis(const Vector2 &direction, float &outMin, float &outMax) const {
    outMin = Vector2::Dot(direction, A);
    outMax = Vector2::Dot(direction, B);

    if (outMax < outMin)
        Swap(outMin, outMax);
}
